/*  display_pup.c -- Display a pup in a popup window.
 *  Copyright (C) 2008-2010  Nick Gasson
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>

#include <gtk/gtk.h>
#include <gtk/gtkwindow.h>

#ifdef WITH_DBUS
#include <dbus/dbus-glib-bindings.h>
#define XPUPSAY_PATH "/uk/me/doof/Pupsay"
#define XPUPSAY_NAMESPACE "uk.me.doof.Pupsay"
#endif

#include "floating_shape.h"
#include "display_pup.h"
#include "settings.h"
#include "i18n.h"

GdkPixbuf *make_text_bubble(char *text, int *p_width, int *p_height, pupmode_t mode);
GdkPixbuf *make_dream_bubble(const char *file, int *p_width, int *p_height);

#define TICK_TIMEOUT   100

#define max(a, b) ((a) > (b) ? (a) : (b))

typedef enum {
   csLeadIn, csDisplay, csLeadOut, csCleanup
} pupstate_t;

typedef struct {
   float_shape_t *pup, *bubble;
   int bubble_width, bubble_height;
   GdkPixbuf *pup_pixbuf, *bubble_pixbuf;
   pupstate_t state;
   int transition_timeout;
   int display_time;
} xpupsay_t;

static xpupsay_t xpupsay;

static gboolean tick(gpointer data);

static pupstate_t next_state(pupstate_t state)
{
   switch (state) {
   case csLeadIn:
      return csDisplay;
   case csDisplay:
      return csLeadOut;
   case csLeadOut:
      return csCleanup;
   case csCleanup:
   default:
      return csCleanup;
   }
}

static GdkPixbuf *load_pup()
{
   char *pup_path;

   const char *alt_image = get_string_option("alt_image");

   if (*alt_image)
      pup_path = strdup(alt_image);
   else
      asprintf(&pup_path, "%s/%s_%s.png", DATADIR,
         get_string_option("image_base"),
         get_string_option("pup_size"));
   
   GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(pup_path, NULL);
   if (NULL == pixbuf) {
      fprintf(stderr, i18n("Failed to load pup image: %s\n"), pup_path);
      exit(EXIT_FAILURE);
   }

   free(pup_path);   
   return pixbuf;
}

static gboolean pup_clicked(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
   if (csDisplay == xpupsay.state) {
      xpupsay.transition_timeout = 0;
      tick(NULL);
   }
   return true;
}

/*
 * Set up a shape to call pup_clicked when it's clicked.
 */
static void close_when_clicked(float_shape_t *shape)
{
   GdkEventMask events = gdk_window_get_events(shape_window(shape)->window);
   events |= GDK_BUTTON_PRESS_MASK;
   gdk_window_set_events(shape_window(shape)->window, events);
   g_signal_connect(G_OBJECT(shape_window(shape)), "button-press-event",
                    G_CALLBACK(pup_clicked), NULL);
}

static gboolean tick(gpointer data)
{
   xpupsay.transition_timeout -= TICK_TIMEOUT;
   if (xpupsay.transition_timeout <= 0) {
      xpupsay.state = next_state(xpupsay.state);
      switch (xpupsay.state) {
      case csLeadIn:
         fprintf(stderr, "Internal Error: Invalid state csLeadIn\n");
         exit(EXIT_FAILURE);
      case csDisplay:
         show_shape(xpupsay.bubble);
         close_when_clicked(xpupsay.bubble);
         xpupsay.transition_timeout = xpupsay.display_time;
         break;
      case csLeadOut:
         hide_shape(xpupsay.bubble);
         xpupsay.transition_timeout = get_int_option("lead_out_time");
         break;
      case csCleanup:
         destroy_shape(xpupsay.pup);
         xpupsay.pup = NULL;
         destroy_shape(xpupsay.bubble);
         xpupsay.bubble = NULL;
         break;
      }
   }
   
   return (xpupsay.state != csCleanup);
}

void pupsay_init(int *argc, char ***argv)
{
   gtk_init(argc, argv);
   
   xpupsay.pup = NULL;
   xpupsay.bubble = NULL;
   xpupsay.bubble_pixbuf = NULL;
   xpupsay.pup_pixbuf = load_pup();
}

static int count_words(const char *s)
{
   bool last_was_space = false;
   int words;
   for (words = 1; *s; s++) {
      if (isspace(*s) && !last_was_space) {
         words++;
         last_was_space = true;
      }
      else
         last_was_space = false;
   }
   return words;
}

static void normal_setup(const char *text, bool debug, pupmode_t mode)
{
   char *text_copy = strdup(text);
   
   // Trim any trailing newline
   size_t len = strlen(text_copy);
   if ('\n' == text_copy[len-1])
      text_copy[len-1] = '\0';
   
   // Count the words and work out the display time, if neccessary
   xpupsay.display_time = get_int_option("display_time");
   if (xpupsay.display_time < 0) {
      int words = count_words(text_copy);
      xpupsay.display_time = words * get_int_option("reading_speed");
      debug_msg("Calculated display time as %dms from %d words\n",
                xpupsay.display_time, words);
   }
   else {
      debug_msg("Using default display time %dms\n", xpupsay.display_time);
   }

   int min_display = get_int_option("min_display_time");
   int max_display = get_int_option("max_display_time");
   if (xpupsay.display_time < min_display) {
      xpupsay.display_time = min_display;
      debug_msg("Display time too short: clamped to %d\n", min_display);
   }
   else if (xpupsay.display_time > max_display) {
      xpupsay.display_time = max_display;
      debug_msg("Display time too long: clamped to %d\n", max_display);
   }
   
   xpupsay.bubble_pixbuf = make_text_bubble(text_copy, &xpupsay.bubble_width,
                                            &xpupsay.bubble_height, mode);
   free(text_copy);
}

static void dream_setup(const char *file, bool debug)
{
   debug_msg("Dreaming file: %s\n", file);

   xpupsay.display_time = get_int_option("display_time");
   if (xpupsay.display_time < 0)
      xpupsay.display_time = get_int_option("dream_time");

   xpupsay.bubble_pixbuf = make_dream_bubble(file, &xpupsay.bubble_width,
                                             &xpupsay.bubble_height);   
}

void display_pup(bool debug, const char *text, bool run_main, pupmode_t mode)
{
   switch (mode) {
   case PUPMODE_NORMAL:
   case PUPMODE_THINK:
      normal_setup(text, debug, mode);
      break;
   case PUPMODE_DREAM:
      dream_setup(text, debug);
      break;
   default:
      fprintf(stderr, "Error: Unsupported pup mode %d\n", mode);
      exit(1);
   }
   
   g_assert(xpupsay.pup_pixbuf);
   xpupsay.pup = make_shape_from_pixbuf(xpupsay.pup_pixbuf);

   int total_width = shape_width(xpupsay.pup)
      + get_int_option("bubble_x")
      + xpupsay.bubble_width;
   int total_height = max(shape_height(xpupsay.pup), xpupsay.bubble_height);

   int bubble_off = max((xpupsay.bubble_height - shape_height(xpupsay.pup))/2, 0);

   GdkScreen *screen = gdk_screen_get_default();

   gint n_monitors = gdk_screen_get_n_monitors(screen);

   gint pick = get_int_option("monitor");
   if (pick < 0 || pick >= n_monitors)
      pick = rand() % n_monitors;
   
   GdkRectangle geom;
   gdk_screen_get_monitor_geometry(screen, pick, &geom);

   int area_w = geom.width - total_width;
   int area_h = geom.height - total_height;

   // Fit the pup on the screen as best as we can
   // The area can't be be zero or we'd get an FPE
   if (area_w < 1)
      area_w = 1;
   if (area_h < 1)
      area_h = 1;

   int pup_x = get_int_option("at_x");
   if (pup_x < 0)
      pup_x = rand() % area_w;
   else if (pup_x >= area_w)
      pup_x = area_w - 1;
   
   int pup_y = get_int_option("at_y");
   if (pup_y < 0)
      pup_y = rand() % area_h;
   else if (pup_y >= area_h)
      pup_y = area_h - 1;

   move_shape(xpupsay.pup,
      geom.x + pup_x,
      geom.y + bubble_off + pup_y);
   show_shape(xpupsay.pup);

   xpupsay.bubble = make_shape_from_pixbuf(xpupsay.bubble_pixbuf);   
   int bx = shape_x(xpupsay.pup) + shape_width(xpupsay.pup)
      + get_int_option("bubble_x");
   int by = shape_y(xpupsay.pup)
      + (shape_height(xpupsay.pup) - shape_height(xpupsay.bubble))/2
      + get_int_option("bubble_y");
   move_shape(xpupsay.bubble, bx, by);
   
   xpupsay.state = csLeadIn;
   xpupsay.transition_timeout = get_int_option("lead_in_time");
   g_timeout_add(TICK_TIMEOUT, tick, NULL);

   close_when_clicked(xpupsay.pup);
   
   if (run_main)
      gtk_main();

   g_object_unref(xpupsay.bubble_pixbuf);
   xpupsay.bubble_pixbuf = NULL;
}

#ifndef WITH_DBUS

bool try_dbus(bool debug, const char *text, pupmode_t mode)
{
   debug_msg("Skipping DBus (disabled by configure)\n");
   return false;
}

#else 

bool try_dbus(bool debug, const char *text, pupmode_t mode)
{
   DBusGConnection *connection;
   GError *error;
   DBusGProxy *proxy;

   g_type_init();
   error = NULL;
   connection = dbus_g_bus_get(DBUS_BUS_SESSION, &error);
   if (NULL == connection) {
      debug_err("Failed to open connection to bus: %s\n", error->message);
      g_error_free(error);
      return false;
   }

   proxy = dbus_g_proxy_new_for_name(connection, XPUPSAY_NAMESPACE,
                                     XPUPSAY_PATH, XPUPSAY_NAMESPACE);
   g_assert(proxy);

   const char *method = NULL;
   switch (mode) {
   case PUPMODE_NORMAL:
      method = "ShowPup";
      break;
   case PUPMODE_THINK:
      method = "Think";
      break;
   case PUPMODE_DREAM:
      method = "Dream";
      break;
   default:
      g_assert(false);
   }  
   
   error = NULL;
   if (!dbus_g_proxy_call(proxy, method, &error, G_TYPE_STRING, text,
                          G_TYPE_INVALID, G_TYPE_INVALID)) {
      debug_err("ShowPup failed: %s\n", error->message);
      g_error_free(error);
      return false;
   }

   return true;
}

#endif /* #ifndef WITH_DBUS */       

void display_pup_or_invoke_daemon(bool debug, const char *text, pupmode_t mode)
{
   if (!try_dbus(debug, text, mode))
      display_pup(debug, text, true, mode);
}
